# tt_double_cut.rb
#-----------------------------------------------------------------------------
# Version: 0.2.0a
# Compatible: SketchUp 7 PC
#             (Other versions untested)
#-----------------------------------------------------------------------------
#
# Creates a matching cutout to parallel faces of cut-out components and syncs
# them.
#
#-----------------------------------------------------------------------------
#
# Thomas Thomassen
# thomas[at]thomthom[dot]net
#
#-----------------------------------------------------------------------------

require 'sketchup.rb'
#require 'ComponentDefinition-delete.rb' # (?) Delete? If needed, embed.

# --- TODO -------------------------------------------------------------------
# - Catch Make Unique
# - Moving multiple DD doesn't sync all
# - Move+Copy * 3 causes too many copies.
# - Check for Errors and Repair
# - Bugsplat on close / on File->New + File->New

module TT_Double_Cut
	
	### CONSTANTS ### --------------------------------------------------------
	
	# Tool ID constants			# Tool Name
	unless file_loaded?('tt_double_cut.rb')
	T_3D_TEXT            = 21013 # 3DTextTool
	T_ARC                = 21065 # ArcTool
	T_CAMERA_DOLLY       = 10523 # CameraDollyTool
	T_CAMERA_ORBIT       = 10508 # CameraOrbitTool
	T_CAMERA_PAN         = 10525 # CameraPanTool
	T_CAMERA_POSITION    = 21169 # PositionCameraTool
	T_CAMERA_WALK        = 10520 # CameraWalkTool
	T_CAMERA_ZOOM        = 10509 # CameraZoomTool
	T_CAMERA_ZOOM_WINDOW = 10526 # CameraZoomWindowTool
	T_CIRCLE             = 21096 # CircleTool
	T_COMPONENT          = 21013 # ComponentTool
	T_COMPONENT_AXIS     = 21126 # ComponentCSTool
	T_DIMENSION          = 21410 # DimensionTool
	T_ERASE              = 21019 # EraseTool
	T_FREEHAND           = 21031 # FreehandTool
	T_FOLLOWME           = 21525 # ExtrudeTool
	T_MODEL_AXIS         = 21126 # SketchCSTool
	T_MOVE               = 21048 # MoveTool
	T_MEASURE            = 21024 # MeasureTool
	T_OFFSET             = 21100 # OffsetTool
	T_PAINT              = 21074 # PaintTool
	T_PASTE              = 21013 # PasteTool
	T_POLYGON            = 21095 # PolyTool
	T_POSITION_TEXTURE   = 21515 # PositionTextureTool
	T_PUSHPULL           = 21041 # PushPullTool
	T_PROTRACTOR         = 21057 # ProtractorTool
	T_RECTANGLE          = 21094 # RectangleTool
	T_ROTATE             = 21129 # RotateTool
	T_SCALE              = 21236 # ScaleTool
	T_SELECTION          = 21022 # SelectionTool
	T_SECTIONCUT         = 21337 # SectionPlaneTool
	T_LINE               = 21020 # SketchTool
	T_TEXT               = 21405 # TextTool
	end
	
	### MODULE VARIABLES ### -------------------------------------------------
	
	# Hash which links the cutouts with the doublecuts.
	# For each link there's two entries:
	# @links[cutout] = doublecut
	# @links[doublecut] = cutout
	# Whenever you need to get the matching cutout instance you just query the @links hash
	# for the instance you want to match.
	# (?) Don't double the entries; use @links.index(value) instead?
	@links = {}
	# Hash keeping cache info on the definition we monitor.
	# Keys are the definition objects.
	# Values are CacheDefinition objects.
	@cache = {}
	# Variable to keep track over the active paths. We use this to monitor and detect when
	# an instance has been opened and edited.
	@path = []
	@active_path_edited = false
	@path_observer = nil
	# Flag to prevent onActivePathChanged from causing and endless loop.
	@updating = false
	# Drawing tools can be whitelisted as it will trigger the Add/Remove event of the Entities observer
	# Manipulation tools is the ones we need to concern about
	@whitelist = [T_3D_TEXT, T_ARC, T_CAMERA_DOLLY, T_CAMERA_ORBIT, T_CAMERA_PAN, T_CAMERA_POSITION,
					T_CAMERA_WALK, T_CAMERA_ZOOM, T_CAMERA_ZOOM_WINDOW, T_CIRCLE, T_COMPONENT,
					T_DIMENSION, T_ERASE, T_FREEHAND, T_FOLLOWME, T_MODEL_AXIS, T_MEASURE, T_OFFSET,
					T_PAINT, T_PASTE, T_POLYGON, T_POSITION_TEXTURE, T_PROTRACTOR, T_RECTANGLE,
					T_SELECTION, T_SECTIONCUT, T_LINE, T_TEXT]
	# Array list of commands to perform
	@commands = []

	### Accessors methods are located at the bottom of this file.
	
	### METHODS ### ----------------------------------------------------------
	
	# --- Add matching cutouts for all cutout instances --- #
	# Loop though the collection and build a list of all the instance definitions.
	# Then go through each instance and try to place a matching doublecut.
	# Check if all definitions are allready doublecuts. If they are, we assume that the user
	# wants to disable the doublecut feature and we then remove all related doublecuts.
	def self.make_cutouts
		model = Sketchup.active_model
		sel = model.selection
		
		# Find all cutout components in selection
		cutouts = sel.to_a
		cutouts.reject { |e|
			not (e.kind_of?(Sketchup::ComponentInstance) && e.glued_to != nil)
			# Optimize. Find only the defintions.
		}
		
		remove = true
		# Build a list of all definitions
		definitions = []
		cutouts.each { |cutout|
			# Ignore the ones we've allready added and any slave components
			next if definitions.include?(cutout.definition)
			next if cutout.definition.get_attribute('double_cut', 'type') == 'slave'
			definitions << cutout.definition
			# Flag that we have a definition not configured as a doublecutting
			remove = false if cutout.definition.get_attribute('double_cut', 'type') == nil
		}
		
		# Loop through all instances,
		# If the remove flag is true, then all definitions in the selection is allready configured
		# to have doublecuts and we will assume that the user wants to remove this feature from 
		# the definition.
		# If it's false, then some definitions are not set to be double cutting and we'll make all
		# of them double cutting
		definitions.each { |d|
			d.instances.each { |i|
				if remove
					double_cut = @links[i]
					
					puts @links.delete(i)
					puts @links.delete(double_cut)
					
					double_cut.erase!
					i.attribute_dictionaries.delete('double_cut')
				else
					self.place_cutout(i)
				end
			}
			if remove
				d.attribute_dictionaries.delete('double_cut')
				@cache.delete(d)
			end
		}
		
		#Sketchup.active_model.commit_operation
	end
	
	# --- Place an matching cutout --- #
	# This method takes an instance as an argument and checks if a matching doublecut can
	# be placed.  If we haven't created the matching doublecut we do that now.
	def self.place_cutout(cutout)
		# If there allready exist a doublecut for this instance we simply abort.
		if @links[cutout] != nil
			puts "\nplace_cutout: Aborting; allready a link.\n"
			return
		end
		
		model = Sketchup.active_model
		sel = model.selection
			
		result = self.intersect_edges(cutout)
		
		if result == nil
			puts "\nPlace cutout failed. No intersecing.\n"
			# We still want to keep an eye on it incase it should be readjusted to intersect.
			# (!) Are we possibly creating too many observers here? Could this be called many times
			# on the same instance?
			# Yup we do! This must change. Only attach this observer when the instance is first
			# added to the model.
			#cutout.add_observer(CutOutObserver.new)
			return
		end
		
		cut_face = result[0]
		vertices = result[1]
		offset = result[2]
		
		# Get the matching doublecut or create one if we're missing it.
		dcut = model.definitions[cutout.definition.name + '_cutout']
		dcut = self.create_double_cut(cutout, vertices, offset) if dcut == nil
		
		# Ensure the definition has the attribute data required.
		if cutout.definition.attribute_dictionaries['double_cut'] == nil
			cutout.definition.set_attribute('double_cut', 'type', 'master')
			cutout.definition.set_attribute('double_cut', 'offset', offset)
			cutout.definition.set_attribute('double_cut', 'cache_name', cutout.definition.name)
			cutout.definition.entities.add_observer(CutOutContentObserver.new)
			@cache[cutout.definition].name = cutout.definition.name
		end
		
		# To offset the component we create a scale transformation using
		# the offset distance. We use that transformation to size up the
		# glue_face's vector. When we use that vector to create a new
		# transformation, which we cross with the cutout instance's 
		# transformation, we have a transformation which will place the
		# new cutout component correctly offset perpendicular to the
		# original cutout.
		vector_scale = Geom::Transformation.scaling( offset * cutout.transformation.zscale )
		vector = cutout.glued_to.normal.axes.z.transform( vector_scale )
		offset_transformation = Geom::Transformation.new( vector )
		instance_transformation = offset_transformation * cutout.transformation
		
		# Insert double cutout component and glue it to the perpendicular face.
		#instance = model.active_entities.add_instance(dcut, instance_transformation)
		#puts 'Parent Entities:'
		#puts cutout.parent
		instance = cutout.parent.entities.add_instance(dcut, instance_transformation)
		instance.glued_to = cut_face
		instance.locked = true
		
		# (!) Check sucess
		# We add a link attribute to the cutout and doublecut so we know they belong together
		# when we load the model next time. After the model is loaded we use the @links
		# hash for direct access so we don't have to traverse all the instances to find the
		# matching link.
		# Check if entityID is allready used. If so, update the instances which allready uses
		# it with their own entityID so we avoid conflict.
		self.verifyID(cutout.entityID)
		cutout.set_attribute('double_cut', 'link', cutout.entityID)
		instance.set_attribute('double_cut', 'link', cutout.entityID)
		
		# We must monitor the instances for changes. An EntityObserver detects when the instance
		# is moved, rotated or scaled.
		# Add EntitiesObserver to detect content change and rebuild the doublecut.
		cutout.add_observer(CutOutObserver.new)
		#cutout.entities.add_observer(CutOutContentObserver.new) (!) monitor definition
		
		# (!) Validate
		# Add the cutout and doublecut to the @links hash for quick access.
		@links[cutout] = instance
		@links[instance] = cutout
		@cache[cutout.definition].instance_count = cutout.definition.instances.length
	end
	
	# Loops though all double cutting instances and checks that there isn't conflicting link IDs.
	# If there is a conflict, assign new link entityIDs to the offending instances.
	# (?) Can you rely that an entity retains it's entityID throughout the model session? Can this
	# be done on model load instead?
	def self.verifyID(entityID)
		model = Sketchup.active_model
		model.definitions.each { |d|
			next if d.get_attribute('double_cut', 'type') != 'master'
			d.instances.each { |i|
				if i.get_attribute('double_cut', 'link') == entityID
					i.set_attribute('double_cut', 'link', i.entityID)
					@links[i].set_attribute('double_cut', 'link', i.entityID) unless @links[i] == nil
				end
			}
		}
	end
	
	# --- Generate the opposite cutout --- #
	def self.create_double_cut(cutout, vertices, offset, update = nil)
		model = Sketchup.active_model
		self.updating = true
		
		#model.start_operation 'Define Doublecut'
		if update == nil
			puts "\nCreating new doublecut..."
			# Ensure that we recognise the definition as a master cutout.
			# We store the offset so we how where the doublecut needs to be placed when we 
			# reposition it.
			# We also cache the definition name so if the definition is renamed we can also
			# rename the doublecut definition.
			cutout.definition.set_attribute('double_cut', 'type', 'master')
			cutout.definition.set_attribute('double_cut', 'offset', offset)
			cutout.definition.set_attribute('double_cut', 'cache_name', cutout.definition.name)
			
			#@cache[cutout.definition].name = cutout.definition.name
			#@cache[cutout.definition].instance_count = cutout.definition.instances.length
			@cache[cutout.definition] = CacheDefinition.new(cutout.definition)
			
			# Monitor the content for changes
			cutout.definition.entities.add_observer(CutOutContentObserver.new)
			
			# (!)
			#puts '--Attaching Definition Observer'
			#puts cutout.definition
			#puts cutout.definition.name
			#r = cutout.definition.add_observer(CutOutInstanceObserver.new) # (!) Doesn't work!
			#puts r
			
			# Create a new definition
			dcut = model.definitions.add(cutout.definition.name + '_cutout')
			dcut.behavior.cuts_opening = true
			dcut.behavior.is2d = true
			dcut.set_attribute('double_cut', 'type', 'slave')	
		else
			puts '--Updating doublecut...'
			dcut = update
			puts dcut.name
			puts '--Erasing old entities...'
			dcut.entities.clear!
		end
		
		puts 'Adding edges...'
		# Add the edges to create the cutout profile.
		vertices.each { |v|
			dcut.entities.add_edges(v)
		}
		puts 'Hiding edges...'
		# Hide the edges
		dcut.entities.each { |e| e.hidden = true }
		#model.commit_operation
		
		# (!) Validate
		puts "\n"
		self.updating = false
		return dcut
	end
	
	
	def self.intersect_edges(cutout)
		model = Sketchup.active_model
		sel = model.selection
			
		# Ensure the component is glued to a face and not a group or component.
		if cutout.glued_to == nil || !cutout.glued_to.kind_of?(Sketchup::Face)	
			puts "\nNot glued to face. Can't cut\n"
			return nil
		end
		
		glue_face = cutout.glued_to
		t = cutout.transformation
		cut_face = nil
		offset = 0
		
		# Find opposite face
		# Loop through the connected faces and see if any of them intersect with any of
		# the edges of our cutout instance. We only investigate faces that are parallel
		# and facing the opposite direction of our glue_face.
		# If any matches are found we assume this is the correct opposite face and
		# collect all intersecting edges (sets of Point3D) and generate a new doublecut
		# component.
		vertices = []	
		glue_face.all_connected.each { |face|
			# Go through each edge in the component and find which edges
			# lies in the same plane as the parallel face.
			next if not (face.kind_of?(Sketchup::Face) && face.normal.reverse.samedirection?(glue_face.normal))
			
			# We have found a candidate for the opposite face; check if any of our
			# edges intersect.
			cutout.entities.each { |edge|
				next if !edge.kind_of?(Sketchup::Edge)
					
				# Check if the edge is in the face
				vs = [] # Array of edge vertices to build the new cutout component
				edge_on_face = true
				edge.vertices.each { |vertex|
					vp = vertex.position
					p = t * vertex.position # Get its global position
					
					if self.point_on_face?(face, p)
						offset = vp[2]
						vp[2] = 0 # Move vertex Z position to 0
						vs << vp
					else
						edge_on_face = false
					end
				}
				# Add the edge vertices to to main array
				vertices << vs if edge_on_face
			}
			
			# If we found vertices on this face, break out of the loop
			# and continue the operation
			if vertices.length > 0
				cut_face = face
				break
			end
		}
		
		if cut_face == nil
			puts '--> Could not create double cut. Did not find intersecting face'
			# (?) Monitor for changes?
			return nil
		end
		
		return [cut_face, vertices, offset]
	end
	
	
	def self.check_new_instances
		@cache.each { |d, c|
			if d.instances.length > c.instance_count
				puts '--> New Instances'
				# New instances
				d.instances.each { |i|
					if TT_Double_Cut.links[i] == nil
						puts '    * Adding new instance'
						TT_Double_Cut.place_cutout(i) 
					end
				}
			end
		}
	end

	
	# Loop through our @links hash.
	# Only delete doublecuts.
	def self.cleanup_links
		puts "\n=== Cleanup ==="
		puts @links
		puts '-'
		puts @cache
		puts "=== Start ==="
		@links.each_pair { |key, value|
			if key.deleted?
				if value.valid? && value.definition.get_attribute('double_cut', 'type') == 'slave'
					value.erase! if value.valid?
				end
				@links.delete(key)
			elsif value.deleted?
				if key.valid? && key.definition.get_attribute('double_cut', 'type') == 'slave'
					key.erase! if key.valid?
				end
				@links.delete(key) 
			end
		}
		@cache.each { |definition, c|
			c.instance_count = definition.instances.length
		}
		puts @links
		puts '-'
		puts @cache
		puts "=== End ===\n"
	end
	
	
	# Find the matching cutout/doublecut of the given instance.
	# This is used by the Model observer to rebuild the @links hash.
	def self.get_matching(instance)
		model = Sketchup.active_model
		id = instance.get_attribute('double_cut', 'link')
		# Narrow down the list of entities to search by determining the matching
		# component definition.
		type = instance.definition.get_attribute('double_cut', 'type')
		if type == 'master'
			links = model.definitions[instance.definition.name + '_cutout']
		else
			str = instance.definition.name
			links = model.definitions[ str[0, str.length - 7] ]
		end
		# Now that we know the doublecut's definition, check it's instances for matching
		# link ID.
		links.instances.each { |i|
			if i.get_attribute('double_cut', 'link') == id
				return i
			end
		}
		# Return nil if we didn't find any match.
		return nil
	end
	
	   
	def self.repair
		# (!)
	end
	
	
	def self.lock_doublecuts(lock = true)
		puts 'Locking: ' + lock.to_s
		Sketchup.active_model.definitions.each { |d|
			next if d.get_attribute('double_cut', 'type') != 'slave'
			d.instances.each { |i| i.locked = lock }
		}
	end
	
	
	def self.remove_doublecuts
		Sketchup.active_model.definitions.each { |d|
			type = d.get_attribute('double_cut', 'type')
			if type == 'slave'
				d.instances.each {|i| i.erase! }
			elsif type == 'master'
				d.instances.each {|i| i.attribute_dictionaries.delete('double_cut') }
				d.attribute_dictionaries.delete('double_cut')
			end			
		}
		@links = {}
	end
	
	
	def self.is_active_path_dc?
		return false if Sketchup.active_model.active_path == nil
		return false if !Sketchup.active_model.active_path.last.kind_of?(Sketchup::ComponentInstance)
		return false if Sketchup.active_model.active_path.last.definition.attribute_dictionaries == nil
		return true if Sketchup.active_model.active_path.last.definition.attribute_dictionaries['double_cut'] != nil
	end

	
	### OBSERVERS ### --------------------------------------------------------
	class CutOutModelObserver < Sketchup::ModelObserver
		def onDeleteModel(model)
			puts 'onDeleteModel'
		end
		
		def onEraseAll(model)
			puts 'onEraseAll'
			# Reset variables
			TT_Double_Cut.links = {}
			TT_Double_Cut.cache = {}
			TT_Double_Cut.path = []
			TT_Double_Cut.active_path_edited = false
			TT_Double_Cut.updating = false
			TT_Double_Cut.commands = []
		end

		def onActivePathChanged(model)
			#puts 'Active Path Changed'
			# (!) When we exit an path we check the last path node for changes if it's an double
			# cutting component. Cache the path to figure out when we exit.
			# Either allways check for changes to double cutting components after they have been
			# exited. Or use the EntitiesObserver to detect add and remove and the ToolsObserver for
			# any other changes. The first might be the safest.
			# (!) Handle nesting
			puts "\n\nonActivePathChanged: " + model.active_path.inspect
			
			# Remove the observer of the previous path and add an observer to the new one.
			puts TT_Double_Cut.path.inspect
			if TT_Double_Cut.path == nil || TT_Double_Cut.path.last == nil
				# When the last path was the model itself the path will be an empty array.
				# We must catch this.
				r = model.entities.remove_observer(TT_Double_Cut.path_observer)
				puts 'Removing old observer from Model: ' + r.to_s
			else
				puts TT_Double_Cut.path.last.entities
				r = TT_Double_Cut.path.last.entities.remove_observer(TT_Double_Cut.path_observer)
				puts 'Removing old observer: ' + r.to_s
			end
			TT_Double_Cut.path_observer = CutOutPathChangeObserver.new
			r = model.active_entities.add_observer(TT_Double_Cut.path_observer)
			puts 'Adding new observer: ' + r.to_s
			
			
			# Prevent an infinite loop. (Can't remember what caused it.)
			if TT_Double_Cut.updating
				puts "Updating! Cancel event\n\n"
				return 
			end
			
			
			# (!) debug
			puts 'cache path'
			puts model.active_path.inspect
			TT_Double_Cut.path = model.active_path
			puts "\n"
			return # (!)
			
			# Now we check if the last path was a doublecut component and if it's been flagged as
			# possibly changed. If it is, then rebuild the doublecut.
			if model.active_path == nil || (TT_Double_Cut.path != nil && model.active_path.length < TT_Double_Cut.path.length)
				# Close
				puts '-Has it been changed?'
				if TT_Double_Cut.active_path_edited
					# (!) Update
					cutout = TT_Double_Cut.path.last
					puts '-The component (might) have changed. Update...'
					result = TT_Double_Cut.intersect_edges(cutout)
					#puts '-Intersected: ' + result.to_s
					if result == nil
						puts '-Update failed. No intersecing.'
						#return
					else
						puts '-updating...'
						#cut_face = result[0]
						vertices = result[1]
						offset = result[2]
						# update entities
						dcut = TT_Double_Cut.links[cutout].definition # (!) Validate
						puts cutout
						puts cutout.definition.name
						puts dcut
						puts dcut.name
						#sleep(1)
						TT_Double_Cut.create_double_cut(cutout, vertices, offset, dcut)
						puts '-Updated!'
					end
				end
				TT_Double_Cut.active_path_edited = false
			else
				# Open
			end
			puts 'cache path'
			puts model.active_path.inspect
			TT_Double_Cut.path = model.active_path
			puts "\n"
		end
		
		def onTransactionCommit(model)
			# (!) Catch Paste in Place and other events
			# Detect instance count changes
			#puts 'Transaction Commit'
			#TT_Double_Cut.check_new_instances (!) Oddity (infinite loop!) - use Timer?
			
			until TT_Double_Cut.commands.length == 0
				command = TT_Double_Cut.commands.shift
				
				case command[0]
					when 'CutOutDefinitionsObserver_onComponentPropertiesChanged'
						TT_Double_Cut.onComponentPropertiesChanged(command[1], command[2])
					when 'CutOutObserver_onChangeEntity'
						TT_Double_Cut.onChangeEntity(command[1])
					when 'CutOutObserver_onEraseEntity'
						#puts "\nDCC erased"
						TT_Double_Cut.onEraseEntity(command[1])
					when 'CutOutPathChangeObserver_onElementAdded'
						#puts "\nonTransactionCommit - Path Element Added"
						TT_Double_Cut.onPathElementAdded(command[1], command[2])
				end
			end
		end
		
		#def onTransactionStart(model)
			#puts 'onTransactionStart'
		#end
	end
	
	
	class CutOutAppObserver < Sketchup::AppObserver
		
		def initialize
			puts 'AppObserver: Init'
			observe_model(Sketchup.active_model)
		end
		
		def onNewModel(model)
			puts 'OnNewModel'
			observe_model(model)
		end
		
		def onOpenModel(model)
			puts 'onOpenModel'
			observe_model(model)
		end
		
		def observe_model(model)
			# Reset variables
			TT_Double_Cut.links = {}
			TT_Double_Cut.cache = {}
			TT_Double_Cut.path = []
			TT_Double_Cut.active_path_edited = false
			TT_Double_Cut.updating = false
			TT_Double_Cut.commands = []
			# Attach Observers
			model.add_observer(CutOutModelObserver.new)
			model.definitions.add_observer(CutOutDefinitionsObserver.new)
			model.tools.add_observer(CutOutToolsObserver.new)
			
			TT_Double_Cut.path_observer = CutOutPathChangeObserver.new
			r = model.active_entities.add_observer(TT_Double_Cut.path_observer)
			TT_Double_Cut.path = model.active_path
			# Find all cutout components and observe the instances (entity observer)
			# EntityObserver is needed to detect move, rotate and scale actions.
			# EntitiesObserver is needed to detect when the content changes.
			model.definitions.each { |d|
				type = d.get_attribute('double_cut', 'type')
				if type == 'master'
					d.instances.each { |i|
						i.add_observer(CutOutObserver.new)
						# Find matching cutout and add to link hash
						match = TT_Double_Cut.get_matching(i)
						next if match == nil
						
						TT_Double_Cut.links[i] = match
						TT_Double_Cut.links[match] = i
					}
					d.entities.add_observer(CutOutContentObserver.new)
				end
			}
		end
	end
	
	
	class CutOutDefinitionsObserver < Sketchup::DefinitionsObserver
		def onComponentPropertiesChanged(definitions, definition)
			TT_Double_Cut.commands << ['CutOutDefinitionsObserver_onComponentPropertiesChanged', definitions, definition]
		end
		
		# No point with this unless we can remove a definition from the model...
		#def onComponentRemoved(definitions, definition)
		#	return # (!) debug
		#end
		
		# In case doublecuttet instances are imported or added via copy+paste from another file
		# we must catch these as well.
		def onComponentAdded(definitions, definition)
			puts "\n--> New Definition Added:"
			puts "--> " + definition.instances.length.to_s
			puts "--> " + definition.get_attribute('double_cut', 'type').inspect
			puts "\n"
			# (!)
		end
	end

	
	# Detect move, rotate and scale. Sync the doublecut.
	class CutOutObserver < Sketchup::EntityObserver
		
		def onChangeEntity(entity)
			puts 'CutOutObserver.onChangeEntity'
			TT_Double_Cut.commands << ['CutOutObserver_onChangeEntity', entity]
		end
		
		def onEraseEntity(entity)
			TT_Double_Cut.commands << ['CutOutObserver_onEraseEntity', entity]
		end
		
	end
	
	
	class CutOutContentObserver < Sketchup::EntitiesObserver
		#def onContentsModified(entities) # (!) Doesn't trigger
		#	puts 'Content Changed'
		#end
		
		def onElementAdded(entities, entity)
			#puts 'Element Added'
			if entity.kind_of?(Sketchup::Edge) && entities.parent == Sketchup.active_model.active_path.last.definition
				TT_Double_Cut.active_path_edited = true
			end
			# (!) What if something was changed when the component wasn't open?
		end
		
		def onElementRemoved(entities, entity)
			#puts 'Element Removed'
			if entity.kind_of?(Sketchup::Edge) && entities.parent == Sketchup.active_model.active_path.last.definition
				TT_Double_Cut.active_path_edited = true
			end
			# (!) What if something was changed when the component wasn't open?
		end
	end
	
	
	class CutOutPathChangeObserver < Sketchup::EntitiesObserver
		def onElementAdded(entities, entity)
			puts 'CutOutPathChangeObserver.onElementAdded'
			TT_Double_Cut.commands << ['CutOutPathChangeObserver_onElementAdded', entities, entity]
		end
		
		#def onElementRemoved(entities, entity)
		#	# (!) Catch new instances
		#	if entity.kind_of?(Sketchup::ComponentInstance)
		#		type = entity.definition.get_attribute('double_cut', 'type')
		#		if type == 'master' && !TT_Double_Cut.updating # Ensure we're not updating
		#			puts '--> PathChange - Remove'
		#			puts entity
		#			puts "\n"
		#		end
		#	end
		#end
	end
	
	
	class CutOutToolsObserver < Sketchup::ToolsObserver
		def onActiveToolChanged(tools, tool_name, tool_id)
			if TT_Double_Cut.is_active_path_dc? && !TT_Double_Cut.whitelist.include?(tool_id)
				puts "onActiveToolChanged: #{tools.to_s}, #{tool_name.to_s}, #{tool_id.to_s}"
				TT_Double_Cut.active_path_edited = true
			end
		end
		
		def onToolStateChanged(tools, tool_name, tool_id, tool_state)
			if TT_Double_Cut.is_active_path_dc? && !TT_Double_Cut.whitelist.include?(tool_id)
				puts "onToolStateChanged: #{tools.to_s}, #{tool_name.to_s}, #{tool_id.to_s}, #{tool_state.to_s}"
			end
			
			#if tool_id == TT_Double_Cut::T_MOVE || tool_id == TT_Double_Cut::T_ROTATE
			#	TT_Double_Cut.check_new_instances
			#end
		end
	end
	
	
	# --- COMMANDS --- #
	# Moved from observer events. Are now called from onTransactionCommit
	def self.onComponentPropertiesChanged(definitions, definition)
		puts definition
		# We syncronise the doublecut names to allways match the names of the 
		# coutouts with "_cutout" appended. That way it's easy to visually see which
		# doublecut belongs to which cutout.
		type = definition.get_attribute('double_cut', 'type')
		if type == 'master'
			puts "\nProperties Changed:"
			puts definition.name
			# Get the cached name which will be the name before it was renamed
			name = definition.get_attribute('double_cut', 'cache_name')
			puts 'Cache name: ' + name.to_s
			if name == nil
				puts 'No cache_name attribute.'
				return
			end
			# Check if the name has changed
			if name == definition.name
				puts 'Name has not changed.'
				return
			end
			# (!) Detect if there's an component with the old cached name. If there is, then
			# the component has been made unique. We must then replace the current doublecut
			# with a new matching one.
			old_def = Sketchup.active_model.definitions[name]
			if old_def != nil
				puts 'Make Unique!'
				puts definition.instances.length
				# (!) instances are empty at this points...
				definition.instances.each { |i|
					
					old_cut = TT_Double_Cut.links[i]
					old_cut.erase!
					
					TT_Double_Cut.place_cutout(entity)
				}
				
				puts "\nOld Def: " + old_def.instances.length.to_s
				old_def.instances.each { |i|
					puts i.inspect + ' - ' + i.definition.inspect
				}
				puts "\n"
				return
			end
			
			# Find the matching doublecut definition.
			match = Sketchup.active_model.definitions[name + '_cutout']
			if match == nil
				puts 'Could not rename match.'
				return
			end
			# Verify renaming
			new_name = definition.name + '_cutout'
			if Sketchup.active_model.definitions.unique_name(new_name) != new_name
				puts 'Error! The new name was not unique.'
			end
			# Rename it and update the cutout's cache_name
			match.name = definition.name + '_cutout'
			definition.set_attribute('double_cut', 'cache_name', definition.name)
		end
	end
	
	
	def self.onChangeEntity(entity)
		#puts 'Deleted: ' + entity.deleted?.to_s
		#puts 'Valid: ' + entity.valid?.to_s
		#return if entity.deleted?
		puts "\n== DCC Entity Changed =="
		return if entity.definition == nil
		
		#Sketchup.active_model.start_operation 'TTDC - onChangeEntity', true, false, true
		
		#puts "\n== Entity Changed =="
		
		type = entity.definition.get_attribute('double_cut', 'type')
		puts '-> ' + type.to_s + ' Cutout Changed: ' + entity.inspect
		
		# (!) Sync the cutout only if the master has moved or scaled.
		
		if type == 'master'
			offset = entity.definition.get_attribute('double_cut', 'offset')
			
			dcut = TT_Double_Cut.links[entity]
			puts dcut
			
			if dcut == nil
				puts '-> No Link Found'
				# Try to place a new cutout
				TT_Double_Cut.place_cutout(entity)
				#Sketchup.active_model.commit_operation 
				return
			end
			puts '-> IDs: ' + entity.get_attribute('double_cut', 'link').to_s + ' -> ' + dcut.get_attribute('double_cut', 'link').to_s
			
			# --- Sync the cutout transformation ---
			# Get a point for which we can create a transformation to move the doublecut
			# in the correct position.
			p = Geom::Point3d.new(0,0,0)
			v = entity.glued_to.normal
			p.offset!(v, offset * entity.transformation.zscale)
			# The distance and direction of this point is the same as the relative distance and
			# offset we want our doublecut to be at.
			
			# No we make a transformation out of it. We must also apply the same transformation
			# as our cutout instance.
			to = Geom::Transformation.new(p)
			tr = to * entity.transformation
			
			# Insertion point which we can test if it's placed properly
			checkpoint = Geom::Point3d.new(0,0,0).transform(tr)
			
			# --- DEBUG --- #
			# Now test the point to see if it's on the face
			classified = dcut.glued_to.classify_point(checkpoint)
			puts "-> Point Classification: #{classified}"
			#dcut.parent.entities.add_cpoint(checkpoint) # Global
			#dcut.parent.entities.add_cpoint(p) # Local
			# --- /DEBUG --- #
			
			if TT_Double_Cut.point_on_face?(dcut.glued_to, checkpoint)
				# Update position
				puts '-> Transform'
				dcut.transformation = tr # (!) This prevents other onEntityChanged events to
				# trigger on the same Move command. Need a workaround.
			else
				puts '-> Remove DD'
				# Doesn't fit on face, remove the doublecut.
				# (!) Move this into a method for later reuse
				##Sketchup.active_model.selection.add dcut
				TT_Double_Cut.links.delete(entity)
				TT_Double_Cut.links.delete(dcut)
				dcut.erase!
				# (?) Try to add a new cutout incase the opposite face is split?
				# Try to place only if it's on the same plane.
				TT_Double_Cut.place_cutout(entity) if classified == 8
			end
			
			# Match layers and visibility
			dcut.layer = entity.layer if dcut.layer != entity.layer
			dcut.hidden = entity.visible? if dcut.hidden? != entity.hidden?
		end
		
		#Sketchup.active_model.commit_operation 
		puts '-> Entity Changed Complete!'
	end
	
	def self.onEraseEntity(entity)
			# entity doesn't appear to return the proper entity that's deleted. It returns no
			# parent nor any attribute dictionaries. The entity object can't be used to lookup
			# the @links hash either. So we have to use the .cleanup_links method as an alternative.
			TT_Double_Cut.cleanup_links
	end
	
	def self.onPathElementAdded(entities, entity)
		# (!) Catch new instances
		if entity.kind_of?(Sketchup::ComponentInstance)
			#puts '--> PathChange - Add'
			#puts entity
			
			# If the component is a cutout with doublecut activated; place a matching
			# doublecut.
			type = entity.definition.get_attribute('double_cut', 'type')
			if type == 'master' && !TT_Double_Cut.updating # Ensure we're not updating
				puts '--> PathChange - Add'
				puts entity
				puts "\n"
				
				TT_Double_Cut.place_cutout(entity)
				# If the component doesn't intersect when placed it won't be monitored.
				# We still want to keep an eye on it incase it should be readjusted to intersect.
				entity.add_observer(CutOutObserver.new) if TT_Double_Cut.links[entity] == nil
			end
		end
		#TT_Double_Cut.check_new_instances
	end
	
	
	### CUSTOM CLASSES ### ---------------------------------------------------
	class CacheDefinition
		attr_accessor(:name, :instance_count)
		def initialize(definition)
			@name = definition.name
			@instance_count = definition.instances.length
		end
	end


	### VALIDATIONS ### ------------------------------------------------------
	def self.selected_cutout_component?
		# Check if the selection contains cutout components
		sel = Sketchup.active_model.selection
		sel.each { |e|
			return true if e.kind_of?(Sketchup::ComponentInstance) && e.glued_to != nil
		}
		return false
	end
	
	def self.menu_validate_cutout
		menu = MF_CHECKED
		# Check if the selection contains cutout components
		sel = Sketchup.active_model.selection
		sel.each { |e|
			if not (e.kind_of?(Sketchup::ComponentInstance) && e.definition.get_attribute('double_cut', 'type') == 'master')
				menu = MF_UNCHECKED
			end
		}
		return menu
	end
	
	def self.point_on_face?(face, point)
		result = face.classify_point(point)
		return true if result > 0 && result < 8
		return false
	end
	
	
	### ACCESSORS ### --------------------------------------------------------
	def self.links
		@links
	end
	def self.links=(value)
		@links = value
	end

	def self.cache
		@cache
	end
	def self.cache=(value)
		@cache = value
	end

	def self.path
		@path
	end
	def self.path=(value)
		@path = value
	end
	
	def self.active_path_edited
		@active_path_edited
	end
	def self.active_path_edited=(value)
		@active_path_edited = value
	end
	
	def self.path_observer
		@path_observer
	end
	def self.path_observer=(value)
		@path_observer = value
	end
	
	def self.updating
		@updating
	end
	def self.updating=(value)
		@updating = value
	end
	
	def self.whitelist
		@whitelist
	end
	def self.whitelist=(value)
		@whitelist = value
	end
	
	def self.commands
		@commands
	end
	def self.commands=(value)
		@commands = value
	end
end


#-----------------------------------------------------------------------------
# Add some menu items to access this
unless file_loaded?('tt_double_cut.rb')
	cmd_create_double_cut = UI::Command.new('Double Cut') { TT_Double_Cut.make_cutouts }
	
	tools_menu = UI.menu('Tools')
    dcut_menu = tools_menu.add_submenu('Doublecuts')
    dcut_menu.add_item('Check for Errors and Repair') { TT_Double_Cut::repair }
    dcut_menu.add_separator()
	dcut_menu.add_item('Lock Doublecuts')	{ TT_Double_Cut::lock_doublecuts(true) }
	dcut_menu.add_item('Unlock Doublecuts')	{ TT_Double_Cut::lock_doublecuts(false) }
	dcut_menu.add_separator()
	dcut_menu.add_item('Remove Doublecuts')	{ TT_Double_Cut::remove_doublecuts }
	
	
	UI.add_context_menu_handler do |menu|
		if TT_Double_Cut.selected_cutout_component?
			item = menu.add_item(cmd_create_double_cut)
			menu.set_validation_proc(item)  { TT_Double_Cut.menu_validate_cutout }
		end
	end
	
	Sketchup.add_observer(TT_Double_Cut::CutOutAppObserver.new)
end
#-----------------------------------------------------------------------------
file_loaded('tt_double_cut.rb')
#-----------------------------------------------------------------------------